/*
Copyright 1995 Massachusetts Institute of Technology.  All Rights 
Reserved.

You are hereby granted a worldwide, irrevocable, paid-up, right and 
license to use, execute, display, modify, copy and distribute MIT's 
Modifications, provided that (i) you abide by the terms and conditions 
of your Transarc AFS License Agreement, and (ii) you do not use the name 
of MIT in any advertising or publicity without the prior written consent 
of MIT.  MIT disclaims all liability for your use of MIT's 
Modifications.  MIT's Modifications are provided "AS IS" WITHOUT 
WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO, 
ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
NONINFRINGEMENT.
*/

/*
 * NetBSD specific assistance routines & VFS ops
 *
 * $Id: afs_nbsd_subr.c,v 1.1.4.4 2000/06/21 12:43:12 sommerfeld Exp $
 */
#ifndef lint
static char vcid[] = "$Id: afs_nbsd_subr.c,v 1.1.4.4 2000/06/21 12:43:12 sommerfeld Exp $";
#endif /* lint */
/*
 * Some code cribbed from ffs_vfsops and other NetBSD sources, which
 * are marked:
 */
/*
 * Copyright (c) 1989, 1991, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <afs/param.h>	/* Should be always first */
#ifdef	AFS_NETBSD_ENV
#include <afs/sysincludes.h>		/* Standard vendor system headers */
#include <afs/afsincludes.h>		/* Afs-based standard headers */
#include <afs/afs_stats.h>		/* statistics */
#include <sys/malloc.h>
#include <sys/namei.h>
#include <sys/conf.h>
#include <sys/exec.h>
#include <sys/lkm.h>
#include <sys/syscall.h>
#include <sys/syscallargs.h>

#define NBSD_DONTFOLLOW_LINK 0
#define NBSD_FOLLOW_LINK 1
static int lkmid = -1;
static int afs_badcall(struct proc *p, void *xx, register_t *yy);

#ifdef __NetBSD_Version__

#define Majorscale 100000000
#define Minorscale 1000000
#define Releasescale 10000
#define Patchscale 100

#define __NetBSD_Version_Major__ (__NetBSD_Version__ / Majorscale )
#define __NetBSD_Version_Minor__ (__NetBSD_Version__ % Majorscale ) / Minorscale
#define __NetBSD_Version_Release__ (__NetBSD_Version__ % Minorscale ) / Releasescale
#define __NetBSD_Version_Patch__ (__NetBSD_Version__ % Releasescale ) / Patchscale

char afs_NetBSD_release[] = {
	'0' + __NetBSD_Version_Major__,
	'.',
	'0' + __NetBSD_Version_Minor__,
#if __NetBSD_Version_Release__ != 0	
	'A' + __NetBSD_Version_Release__ - 1,
#endif
#if __NetBSD_Version_Patch__ != 0
	'.',
	'0' + __NetBSD_Version_Patch__,
#endif
	0,
	'P', 'a', 't', 'c', 'h', ' ', 'f', 'o', 'd', 'd', 'e', 'r',
	0,
};
#else
#if NetBSD1_3 == 1
char afs_NetBSD_release[] = "1.3\0Patch fodder";
#elif NetBSD1_3 == 2
char afs_NetBSD_release[] = "1.3A\0Patch fodder";
#elif NetBSD1_3 == 3
char afs_NetBSD_release[] = "1.3B\0Patch fodder";
#elif NetBSD1_3 == 4
char afs_NetBSD_release[] = "1.3C\0Patch fodder";
#elif NetBSD1_3 == 5
char afs_NetBSD_release[] = "1.3D\0Patch fodder";
#elif NetBSD1_3 == 6
char afs_NetBSD_release[] = "1.3E\0Patch fodder";
#elif NetBSD1_3 == 7
char afs_NetBSD_release[] = "1.3F\0Patch fodder";
#elif NetBSD1_3 == 8
char afs_NetBSD_release[] = "1.3G\0Patch fodder";
#elif NetBSD1_3 == 9
char afs_NetBSD_release[] = "1.3H\0Patch fodder";
#elif NetBSD1_3 == 10
char afs_NetBSD_release[] = "1.3I\0Patch fodder";
#elif NetBSD1_3 == 11
char afs_NetBSD_release[] = "1.3J\0Patch fodder";
#elif NetBSD1_3 == 12
char afs_NetBSD_release[] = "1.3K\0Patch fodder";
#elif NetBSD1_3 == 13
char afs_NetBSD_release[] = "1.3L\0Patch fodder";
#elif NetBSD1_2 == 1
#if defined(SS_PRIV)
char afs_NetBSD_release[] = "1.2\0Patch fodder";
#else
char afs_NetBSD_release[] = "1.2.1\0Patch fodder";
#endif
#elif NetBSD1_2 == 2
char afs_NetBSD_release[] = "1.2A\0Patch fodder";
#elif NetBSD1_2 == 3
char afs_NetBSD_release[] = "1.2B\0Patch fodder";
#elif NetBSD1_2 == 4
char afs_NetBSD_release[] = "1.2C\0Patch fodder";
#elif NetBSD1_2 == 5
char afs_NetBSD_release[] = "1.2D\0Patch fodder";
#elif NetBSD1_2 == 6
char afs_NetBSD_release[] = "1.2E\0Patch fodder";
#elif NetBSD1_2 == 7
char afs_NetBSD_release[] = "1.2F\0Patch fodder";
#elif NetBSD1_2 == 8
char afs_NetBSD_release[] = "1.2G\0Patch fodder";
#elif NetBSD1_2 == 9
char afs_NetBSD_release[] = "1.2H\0Patch fodder";
#elif NetBSD1_2 == 10
char afs_NetBSD_release[] = "1.2I\0Patch fodder";
#elif NetBSD1_1 == 1
char afs_NetBSD_release[] = "1.1\0Patch fodder";
#elif NetBSD1_1 == 2
char afs_NetBSD_release[] = "1.1A\0Patch fodder";
#elif NetBSD1_1 == 3
char afs_NetBSD_release[] = "1.1B\0Patch fodder";
#endif

#endif

char afs_NetBSD_osname[] = "NetBSD";

int
afs_nbsd_lookupname(fnamep, segflg, followlink, dirvpp, compvpp)
	char *fnamep;
	enum uio_seg segflg;
	int followlink;
	struct vnode **dirvpp;
	struct vnode **compvpp;
{
    struct nameidata nd;
    int niflag;
    int error;

    /*
     * Lookup pathname "fnamep", returning parent directory in
     * *dirvpp (if non-null) and leaf in *compvpp.  segflg says whether the
     * pathname is user or system space.
     */
    /* XXX LOCKLEAF ? */
    niflag =  (followlink == NBSD_FOLLOW_LINK) ? FOLLOW : NOFOLLOW;
    if (dirvpp)
	niflag |= WANTPARENT;		/* XXX LOCKPARENT? */
    NDINIT(&nd, LOOKUP,
	   niflag,
	   segflg,
	   fnamep, osi_curproc());
    if (error = namei(&nd))
	return error;
    *compvpp = nd.ni_vp;
    if (dirvpp)
	*dirvpp = nd.ni_dvp;
    return error;
}

extern int
afs_nbsd_rdwr(vp, uiop, op, flags, cred)
	struct vnode *vp;
	struct uio *uiop;
	enum uio_rw op;
	int flags;
	struct AFS_UCRED *cred;
{
    uiop->uio_rw = op;
    if (op == UIO_READ)
	return VOP_READ(vp, uiop, flags, cred);
    if (op == UIO_WRITE)
	return VOP_WRITE(vp, uiop, flags, cred);
#ifdef DIAGNOSTIC
    panic("afs_nbsd_rdwr mode");
#endif
    return EINVAL;
}

int
afs_nbsd_quotactl()
{
	return EOPNOTSUPP;
}

int
afs_nbsd_fhtovp(mp, fhp, vpp)
	struct mount *mp;
	struct fid *fhp;
	struct vnode **vpp;
{

	return (EINVAL);
}

int
afs_nbsd_vptofh(vp, fhp)
	struct vnode *vp;
	struct fid *fhp;
{

	return (EINVAL);
}

#if __NetBSD_Version__ - 0 >= 103050000
int     
afs_nbsd_sysctl(name, namelen, oldp, oldlenp, newp, newlen, p)
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	struct proc *p;
{     
	return (EOPNOTSUPP);
}
#endif

#if __NetBSD_Version__ - 0 >= 103100000
int     
afs_nbsd_checkexp(mp, nam, exflagsp, credanonp)
	register struct mount *mp;
	struct mbuf *nam;
	int *exflagsp;
	struct ucred **credanonp;
{
	return (EACCES);
}
#endif

static char afsgenmem[] = "afsgenmem";
static char afsfidmem[] = "afsfidmem";
static char afsbhdrmem[] = "afsbhdrmem";
static char afsbfrmem[] = "afsbfrmem";

static int afs_nbsd_inited = 0;

extern struct vnodeopv_desc afs_nbsd_vnodeop_opv_desc;
extern char *memname[];

static void afs_nbsd_initialize __P((void));

static void
afs_nbsd_initialize()
{

    if (!afs_nbsd_inited) {
#if __NetBSD_Version__ - 0 < 103040000
	vfs_opv_init_explicit(&afs_nbsd_vnodeop_opv_desc);
	vfs_opv_init_default(&afs_nbsd_vnodeop_opv_desc);
#endif
	printf("AFS 3.3a for %s %s initialized\n",
	    afs_NetBSD_osname, afs_NetBSD_release);
	afs_nbsd_inited = 1;
	if (memname[M_AFSGENERIC] == NULL)
	    memname[M_AFSGENERIC] = afsgenmem;
	if (memname[M_AFSFID] == NULL)
	    memname[M_AFSFID] = afsfidmem;
	if (memname[M_AFSBUFHDR] == NULL)
	    memname[M_AFSBUFHDR] = afsbhdrmem;
	if (memname[M_AFSBUFFER] == NULL)
	    memname[M_AFSBUFFER] = afsbfrmem;
    }
}

int
afs_nbsd_start(mp, flags, p)
	struct mount *mp;
	int flags;
	struct proc *p;
{

    afs_nbsd_initialize();
    return (0);				/* nothing else to do. */
}

int
afs_nbsd_mount(mp, path, data, ndp, p)
	register struct mount *mp;
	char *path;
	caddr_t data;
	struct nameidata *ndp;
	struct proc *p;
{
    /* ndp contains the mounted-from device.  Just ignore it.
       we also don't care about our proc struct. */
    int size;
    int error;
    extern int no_going_back;

    if (mp->mnt_flag & MNT_UPDATE)
	return EINVAL;

    if (no_going_back) {
	printf("AFS: mount rejected\n");
	printf("AFS: you must unload the module and reload to restart AFS.\n");
	printf("AFS: Run `/sbin/modunload -i %d' to unload AFS.\n", lkmid);
	return EINVAL;
    }

    error = afs_mount(mp, path, data);
    
    if (!error) {
	(void) copyinstr(path, mp->mnt_stat.f_mntonname, MNAMELEN-1, &size);
	bzero(mp->mnt_stat.f_mntonname + size, MNAMELEN - size);
	bzero(mp->mnt_stat.f_mntfromname, MNAMELEN);
	strcpy(mp->mnt_stat.f_mntfromname, "AFS");
	/* null terminated string "AFS" will fit, just leave it be. */
	strcpy(mp->mnt_stat.f_fstypename, MOUNT_AFS);
	(void) afs_statfs(mp, &mp->mnt_stat);
    }
    return error;
}

int
afs_nbsd_unmount(afsp, flags, p)
	struct mount *afsp;
	int flags;
	struct proc *p;
{
    int err;
    extern int sys_ioctl(), sys_setgroups();

    err = afs_unmount(afsp, flags);
    if (err == 0) {
	/* give up syscall entries for ioctl & setgroups, which we've stolen */
	sysent[SYS_ioctl].sy_call = sys_ioctl;
	sysent[SYS_setgroups].sy_call = sys_setgroups;
	/* give up the stolen syscall entry */
	sysent[AFS_SYSCALL].sy_narg = 0;
	sysent[AFS_SYSCALL].sy_argsize = 0;
	sysent[AFS_SYSCALL].sy_call = afs_badcall;
	printf("AFS unmounted--use `/sbin/modunload -i %d' to unload before restarting AFS\n", lkmid);
    }
    return err;
}

static int
afs_badcall(p, xx, yy)
	struct proc *p;
	void *xx;
	register_t *yy;
{
    return ENOSYS;
}

extern struct osi_vfs *afs_globalVFS;

void
afs_nbsd_getnewvnode(tvc)
	struct vcache *tvc;
{
    while (getnewvnode(VT_AFS, afs_globalVFS, afs_nbsd_vnodeop_p, &tvc->v)) {
	/* no vnodes available, force an alloc (limits be damned)! */
	desiredvnodes++;
    }
    tvc->v->v_data = (void *)tvc;
}

int
afs_nbsd_root(mp, vpp)
	struct mount *mp;
	struct vnode **vpp;
{
    int error;
    error = afs_root(mp, vpp);
    if (!error)
#if __NetBSD_Version__ - 0 >= 103050000
	vn_lock(*vpp, LK_EXCLUSIVE | LK_RETRY);
#else
	VOP_LOCK(*vpp);			/* return it locked */
#endif
    return error;
}

void
afs_nbsd_ref(vp)
	struct vnode *vp;
{
    if (vp->v_usecount == 0) {
	vprint("holding unheld node", vp);
	panic("afs_nbsd_ref");
    }
    VREF(vp);
}

void
afs_nbsd_rele(vp)
	struct vnode *vp;
{
    if (vp->v_usecount <= 0) {
	vprint("rele'ing unheld node", vp);
	panic("afs_nbsd_rele");
    }
    vrele(vp);
}

int
afs_nbsd_vget(vp, lfl)
	struct vnode *vp;
	int lfl;
{
    int error;
    if (vp->v_usecount < 0) {
	vprint("bad usecount", vp);
	panic("afs_nbsd_vget");
    }
#if __NetBSD_Version__ - 0 >= 103050000
    error = vget(vp, lfl ? LK_EXCLUSIVE : 0);
#else
    error = vget(vp, lfl);
#endif
    if (!error)
	insmntque(vp, afs_globalVFS);	/* take off free list */
    return error;
}

extern struct vfsops afs_vfsops;

MOD_VFS("afs", 0, &afs_vfsops);

int
afs_vfs_load(lkmtp, cmd)
	struct lkm_table *lkmtp;
	int cmd;
{
   
    afs_nbsd_initialize();
    lkmid = lkmtp->id;
    return 0;
}

int
afs_vfs_unload(lkmtp, cmd)
	struct lkm_table *lkmtp;
	int cmd;
{
    extern struct vcache *afs_globalVp;
    extern int sys_lkmnosys();

    if (afs_globalVp)
	return EBUSY;
    if (sysent[SYS_ioctl].sy_call != sys_ioctl)
    	return EBUSY;

    if (memname[M_AFSGENERIC] == afsgenmem)
	memname[M_AFSGENERIC] = NULL;
    if (memname[M_AFSFID] == afsfidmem)
	memname[M_AFSFID] = NULL;
    if (memname[M_AFSBUFHDR] == afsbhdrmem)
	memname[M_AFSBUFHDR] = NULL;
    if (memname[M_AFSBUFFER] == afsbfrmem)
	memname[M_AFSBUFFER] = NULL;

    sysent[AFS_SYSCALL].sy_call = sys_lkmnosys;
    printf("AFS 3.3a unloaded\n");
    return 0;
}


/*
 * compatibility matrix:
 *
 * build on x.yQ, run only on x.yQ
 * build on x.y, run on x.y.z
 * build on x.y.z, run on x.y.zQ
 */
 


int
afsmodload(lkmtp, cmd, ver)
	struct lkm_table *lkmtp;
	int cmd;
	int ver;
{
	extern int sys_lkmnosys();
	extern char osrelease[], ostype[];
	int buildrellen = strlen(afs_NetBSD_release);
	int osrellen = strlen(osrelease);
	int fail = 0;
    
	if (cmd == LKM_E_LOAD) {
		if (strcmp(ostype,afs_NetBSD_osname)) {
			fail++;
		} else if (strcmp(osrelease,afs_NetBSD_release)) {
#if __NetBSD_Version_Release__ == 0
			if (osrellen < buildrellen)
				fail++;
			else if (strncmp(osrelease,afs_NetBSD_release,buildrellen))
				fail++;
			else if (osrellen < (buildrellen + 1))
				fail++;
#if __NetBSD_Version_Patch__ == 0 
			else if (osrelease[buildrellen] != '.')
				fail++;
#endif
#endif
		}
		if (fail) {
			printf("Error: This version of AFS was built for"
			    " %s version %s\n",
			    afs_NetBSD_osname, afs_NetBSD_release);
			if (fail)
				return EINVAL;
		}
		if (sysent[AFS_SYSCALL].sy_call != sys_lkmnosys) {
			printf("AFS must be loaded with syscall %d assigned"
			    " to sys_lkmnosys\nIs AFS already loaded?\n",
			    AFS_SYSCALL);
			return EINVAL;
		}
	}
	DISPATCH(lkmtp,cmd,ver,afs_vfs_load,afs_vfs_unload,lkm_nofunc);
}
#endif /* AFS_NETBSD_ENV */

